#ifndef PTR_PARTAGE_H
#define PTR_PARTAGE_H

#include "Testable.h"
#include <utility> // std::swap()

template <class T>
   class Partageable
   {
   public:
      typedef unsigned long cpt_type;
      typedef cpt_type* ptr_cpt_type;
   private:
      mutable cpt_type* cptref_;
   public:
      typedef typename T value_type;
      typedef typename value_type* pointer;
   private:
      pointer ptr_;
   public:
      Partageable(pointer p = 0)
         : ptr_(p), cptref_(0)
      {
         if (ptr_) create();
      }
      Partageable(const Partageable &p) throw()
         : ptr_(p.get()), cptref_(0)
      {
         p.remap(cptref_);
      }
      void remap(ptr_cpt_type &p) const
         { if (p = cptref_) AddRef(); }
      template <class U>
         Partageable(const Partageable<U> &p) throw()
            : ptr_(0), cptref_(0)
         {
            if(ptr_= dynamic_cast<T*>(p.get()))
               p.remap(cptref_);
         }
      void swap(Partageable &p) throw()
      {
         std::swap(ptr_, p.ptr_);
         std::swap(cptref_, p.cptref_);
      }
      Partageable& operator=
         (const Partageable &p) throw()
      {
         Partageable(p).swap(*this);
         return *this;
      }
      template <class U>
         Partageable& operator=
            (const Partageable<U> &p) throw()
         {
            Partageable(p).swap(*this);
            return *this;
         }
      ~Partageable() throw()
         { if (cptref_) Release(); }
   private:
      void create() const
         { cptref_ = new cpt_type(1); }
      void AddRef() const throw()
         { ++(*cptref_); }
      void Release() const throw()
      {
         if (--(*cptref_) == 0)
         {
            delete cptref_;
            delete ptr_;
         }
      }
   public:
      pointer get() const throw()
         { return ptr_; }
   };

namespace std
{
   template <class T>
      void swap(Partageable<T> &a, Partageable<T> &b) throw()
         { a.swap(b); }
}

template <class T>
   class PtrPartage
      : public Testable
   {
      Partageable<T> ptr_;
   public:
      Partageable<T> partageable() const
         { return ptr_; }
   private:
      class AutoRelacheur
      {
         Partageable <T> ptr_;
      public:
         AutoRelacheur(const Partageable<T> &p)
            : ptr_(p)
         {
         }
      };
      T* get() const throw()
         { return ptr_.get(); }
   public:
      explicit PtrPartage(T *p = 0)
         : ptr_(p)
      {
      }
      PtrPartage(const PtrPartage &p) throw()
         : ptr_(p.partageable())
      {
      }
      template <class U>
         PtrPartage(const PtrPartage<U> &p) throw()
            : ptr_(p.partageable())
         {
         }
      void swap(PtrPartage &p)
         { std::swap(ptr_, p.ptr_); }
      PtrPartage& operator=(const PtrPartage &p)
      {
         PtrPartage(p).swap(*this);
         return *this;
      }
      template <class U>
         PtrPartage& operator=(const PtrPartage<U> &p)
         {
            PtrPartage (p).swap(*this);
            return *this;
         }
      T& operator*()
         { return *(get()); }
      const T& operator*() const // indirection importante si T est abstrait
         { return *(get()); }
      T* operator->()
      {
         AutoRelacheur ar(ptr_);
         return get();
      }
      const T* operator->() const
      {
         AutoRelacheur ar(ptr_);
         return get();
      }
      //
      // RAFFINER
      //
   private:
      bool_type validite() const
      {
         return get()?
            true_value() : false_value();
      }
   };

template <class T, class U>
   PtrPartage<T> partageable_cast(const PtrPartage<U> &p)
   {
      return PtrPartage<T>(p);
   }

namespace std
{
   template <class T>
      void swap(PtrPartage<T> &a, PtrPartage<T> &b) throw()
         { a.swap(b); }
}

#endif

